# Copyright (c) Fireworks AI, Inc. and affiliates.
#
# All Rights Reserved.

import configparser
import os
import subprocess
from typing import Any, Callable, List, Optional
import grpc
from fireworks.client.log import log_warning
from fireworks.protos.generated.gateway.gateway_pb2_grpc import GatewayStub
from fireworks.protos.generated.gateway.model_pb2 import (
    CreateModelRequest,
    DeleteModelRequest,
    GetModelRequest,
    ListModelsRequest,
    Model,
)


def _get_token_from_firectl() -> str:
    """
    Attempts to obtain auth token from firectl.

    Returns:
        token returned by firectl or an empty string if
        missing.
    """
    try:
        result = subprocess.run(["firectl", "token"], capture_output=True, text=True)
    except:
        return ""
    if result.returncode != 0:
        return ""
    return result.stdout


def _get_token_from_env() -> str:
    """
    Attempts to obtain auth token from the environment variable.

    Returns:
        token retrieved from env variable or an empty string if
        missing.
    """
    return os.environ.get("FIREWORKS_API_TOKEN", "")


def _get_token_from_config() -> str:
    """
    Attempts to obtain auth token from the config file.

    Returns:
        token retrieved from the config or an empty string if
        missing.
    """
    path = os.path.expanduser("~/.fireworks/auth.ini")
    try:
        with open(path, "r") as f:
            content = "[dummy_section]\n" + f.read()
        config = configparser.ConfigParser()
        config.read_string(content)
        return config.get("dummy_section", "id_token")
    except:
        return ""


def _call_with_token_refresh(func: Callable) -> Callable:
    """
    If the initial fuction invocation fails with an authentication error,
    attempts to refresh the credentials and retry the call.

    It is meant to be used as a decorator attached to grpc client wrappers.

    Args:
        func: the function performing grpc call.

    Returns:
        wrapper around the input function.
    """

    def _func(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except grpc.RpcError as rpc_error:
            if rpc_error.code() == grpc.StatusCode.UNAUTHENTICATED:
                log_warning("Token is not valid. Will attempt a refresh")
                connect = getattr(args[0], "_connect")
                connect(
                    [
                        _get_token_from_firectl,
                        _get_token_from_config,
                        _get_token_from_env,
                    ]
                )
        return func(*args, **kwargs)

    return _func


class Gateway(GatewayStub):
    """
    Control plane gateway client that exposes its endpoints through
    convenient APIs.

    Keep the API consistent with `gateway.proto`.
    """

    def __init__(self, *, server_addr: str = "gateway.fireworks.ai:443") -> None:
        """
        Args:
            server_addr: the network address of the gateway server.
        """
        self._server_addr = server_addr
        self._connect()

    def _connect(self, sources: Optional[List[Callable]] = None) -> None:
        """
        Obtains credentials and creates a stub connecting to the gateway.

        Args:
            sources: the list of authentication token sources.
        """
        if sources is None:
            sources = [
                _get_token_from_env,
                _get_token_from_config,
                _get_token_from_firectl,
            ]
        token = ""
        for source in sources:
            token = source()
            if token:
                break
        if not token:
            raise RuntimeError(
                "cannot find fireworks API token. To obtain it, install `firectl` "
                "following the instructions at https://fireworks.ai/docs/install/ "
                "and call `firectl signin <FIREWORKS_ACCOUNT_ID>`. Alternatively, "
                "you can set FIREWORKS_API_TOKEN environment variable"
            )
        creds = grpc.composite_channel_credentials(
            grpc.ssl_channel_credentials(), grpc.access_token_call_credentials(token)
        )
        channel = grpc.secure_channel(self._server_addr, creds)
        self._stub = GatewayStub(channel)

    @_call_with_token_refresh
    def create_model(
        self, *, parent: str = "", model: Optional[Model] = None, model_id: str = ""
    ) -> Model:
        """
        Creates a model.

        Args:
            parent: resource name of the parent account,
            model: properties of the model being created,
            model_id: id of the model.

        Returns:
            properties of the created model.
        """
        request = CreateModelRequest()
        request.parent = parent
        request.model = model
        request.model_id = model_id
        return self._stub.CreateModel(request)

    @_call_with_token_refresh
    def get_model(self, *, name: str = "") -> Model:
        """
        Obtains proprties of a model.

        Args:
            name: resource name of the model.

        Returns:
            properties of the model with the given name.
        """
        request = GetModelRequest()
        request.name = name
        return self._stub.GetModel(request)

    @_call_with_token_refresh
    def delete_model(self, *, name: str = "") -> None:
        """
        Deletes a model.

        Args:
            name: resource name of the model to delete.
        """
        request = DeleteModelRequest()
        request.name = name
        self._stub.DeleteModel(request)

    @_call_with_token_refresh
    def list_models(
        self,
        *,
        parent: str = "",
        filter: str = "",
        order_by: str = "",
    ) -> List[Model]:
        """
        Fetches the list of available models.

        Args:
            parent: resource name of the parent account,
            filter: only models satisfying the provided filter (if specified)
                will be returned. See https://google.aip.dev/160 for the filter
                grammar,
            order_by: a comma-separated list of fields to order by. e.g. "foo,bar".
                The default sort order is ascending. To specify a descending order
                for a field, append a " desc" suffix. e.g. "foo desc,bar"
                Subfields are specified with a "." character. e.g. "foo.bar".
                If not specified, the default order is by "name".

        Returns:
            list of models satisfying the retrieval criteria.
        """
        result = []
        page_token = None
        while True:
            request = ListModelsRequest()
            request.parent = parent
            request.filter = filter
            request.order_by = order_by
            if page_token is not None:
                request.page_token = page_token
            response = self._stub.ListModels(request)
            result.extend(response.models)
            if response.total_size < len(result):
                return result
                # FIXME(pawel): there is a bug in the control plane code
                # that causes inconsistency between the total_size and the
                # result size. Uncomment the line below after it gets fixed.
                # raise ValueError(
                #     f"response total size {response.total_size} is lower than the "
                #     f"result length {len(result)}")
            elif response.total_size == len(result):
                return result
            page_token = response.next_page_token



from . import options_pb2 as _options_pb2
from ..google.api import field_behavior_pb2 as _field_behavior_pb2
from ..google.api import resource_pb2 as _resource_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class AbuseRule(_message.Message):
    __slots__ = ("name", "target", "type", "designation", "create_time", "update_time", "reason", "state")
    class Type(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        TYPE_UNSPECIFIED: _ClassVar[AbuseRule.Type]
        DOMAIN: _ClassVar[AbuseRule.Type]
        IP: _ClassVar[AbuseRule.Type]
        ACCOUNT: _ClassVar[AbuseRule.Type]
    TYPE_UNSPECIFIED: AbuseRule.Type
    DOMAIN: AbuseRule.Type
    IP: AbuseRule.Type
    ACCOUNT: AbuseRule.Type
    class State(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        STATE_UNSPECIFIED: _ClassVar[AbuseRule.State]
        CREATING: _ClassVar[AbuseRule.State]
        READY: _ClassVar[AbuseRule.State]
        UPDATING: _ClassVar[AbuseRule.State]
        DELETING: _ClassVar[AbuseRule.State]
    STATE_UNSPECIFIED: AbuseRule.State
    CREATING: AbuseRule.State
    READY: AbuseRule.State
    UPDATING: AbuseRule.State
    DELETING: AbuseRule.State
    class Designation(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        DESIGNATION_UNSPECIFIED: _ClassVar[AbuseRule.Designation]
        ALLOWED: _ClassVar[AbuseRule.Designation]
        BLOCKED: _ClassVar[AbuseRule.Designation]
        SUSPICIOUS: _ClassVar[AbuseRule.Designation]
    DESIGNATION_UNSPECIFIED: AbuseRule.Designation
    ALLOWED: AbuseRule.Designation
    BLOCKED: AbuseRule.Designation
    SUSPICIOUS: AbuseRule.Designation
    NAME_FIELD_NUMBER: _ClassVar[int]
    TARGET_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    DESIGNATION_FIELD_NUMBER: _ClassVar[int]
    CREATE_TIME_FIELD_NUMBER: _ClassVar[int]
    UPDATE_TIME_FIELD_NUMBER: _ClassVar[int]
    REASON_FIELD_NUMBER: _ClassVar[int]
    STATE_FIELD_NUMBER: _ClassVar[int]
    name: str
    target: str
    type: AbuseRule.Type
    designation: AbuseRule.Designation
    create_time: _timestamp_pb2.Timestamp
    update_time: _timestamp_pb2.Timestamp
    reason: str
    state: AbuseRule.State
    def __init__(self, name: _Optional[str] = ..., target: _Optional[str] = ..., type: _Optional[_Union[AbuseRule.Type, str]] = ..., designation: _Optional[_Union[AbuseRule.Designation, str]] = ..., create_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., update_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., reason: _Optional[str] = ..., state: _Optional[_Union[AbuseRule.State, str]] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: gateway/api_key.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'gateway/api_key.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..google.api import field_behavior_pb2 as google_dot_api_dot_field__behavior__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15gateway/api_key.proto\x12\x07gateway\x1a\x1fgoogle/api/field_behavior.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"\xb5\x01\n\x06\x41piKey\x12\x16\n\x06key_id\x18\x01 \x01(\tB\x06\xe2\x41\x03\x01\x03\x05\x12\x1a\n\x0c\x64isplay_name\x18\x02 \x01(\tB\x04\xe2\x41\x01\x01\x12\x13\n\x03key\x18\x03 \x01(\tB\x06\xe2\x41\x03\x01\x03\x05\x12\x36\n\x0b\x63reate_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x05\xe2\x41\x02\x03\x05\x12\x15\n\x06secure\x18\x06 \x01(\x08\x42\x05\xe2\x41\x02\x03\x05\x12\x13\n\x05\x65mail\x18\x07 \x01(\tB\x04\xe2\x41\x01\x03\"S\n\x13\x43reateApiKeyRequest\x12\x14\n\x06parent\x18\x01 \x01(\tB\x04\xe2\x41\x01\x02\x12&\n\x07\x61pi_key\x18\x02 \x01(\x0b\x32\x0f.gateway.ApiKeyB\x04\xe2\x41\x01\x02\"\x8b\x01\n\x12ListApiKeysRequest\x12\x14\n\x06parent\x18\x01 \x01(\tB\x04\xe2\x41\x01\x02\x12\x17\n\tpage_size\x18\x02 \x01(\x05\x42\x04\xe2\x41\x01\x01\x12\x18\n\npage_token\x18\x03 \x01(\tB\x04\xe2\x41\x01\x01\x12\x14\n\x06\x66ilter\x18\x04 \x01(\tB\x04\xe2\x41\x01\x01\x12\x16\n\x08order_by\x18\x05 \x01(\tB\x04\xe2\x41\x01\x01\"e\n\x13ListApiKeysResponse\x12!\n\x08\x61pi_keys\x18\x01 \x03(\x0b\x32\x0f.gateway.ApiKey\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\x12\x12\n\ntotal_size\x18\x03 \x01(\x05\"A\n\x13\x44\x65leteApiKeyRequest\x12\x14\n\x06parent\x18\x01 \x01(\tB\x04\xe2\x41\x01\x02\x12\x14\n\x06key_id\x18\x02 \x01(\tB\x04\xe2\x41\x01\x02\"0\n\x18GetInternalApiKeyRequest\x12\x14\n\x06parent\x18\x01 \x01(\tB\x04\xe2\x41\x01\x02\"R\n\x1aUpdateApiKeyAccountRequest\x12\x19\n\x0bold_account\x18\x01 \x01(\tB\x04\xe2\x41\x01\x02\x12\x19\n\x0bnew_account\x18\x02 \x01(\tB\x04\xe2\x41\x01\x02\x42\x43ZAgithub.com/fw-ai/fireworks/control_plane/protos/generated/gatewayb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'gateway.api_key_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'ZAgithub.com/fw-ai/fireworks/control_plane/protos/generated/gateway'
  _globals['_APIKEY'].fields_by_name['key_id']._loaded_options = None
  _globals['_APIKEY'].fields_by_name['key_id']._serialized_options = b'\342A\003\001\003\005'
  _globals['_APIKEY'].fields_by_name['display_name']._loaded_options = None
  _globals['_APIKEY'].fields_by_name['display_name']._serialized_options = b'\342A\001\001'
  _globals['_APIKEY'].fields_by_name['key']._loaded_options = None
  _globals['_APIKEY'].fields_by_name['key']._serialized_options = b'\342A\003\001\003\005'
  _globals['_APIKEY'].fields_by_name['create_time']._loaded_options = None
  _globals['_APIKEY'].fields_by_name['create_time']._serialized_options = b'\342A\002\003\005'
  _globals['_APIKEY'].fields_by_name['secure']._loaded_options = None
  _globals['_APIKEY'].fields_by_name['secure']._serialized_options = b'\342A\002\003\005'
  _globals['_APIKEY'].fields_by_name['email']._loaded_options = None
  _globals['_APIKEY'].fields_by_name['email']._serialized_options = b'\342A\001\003'
  _globals['_CREATEAPIKEYREQUEST'].fields_by_name['parent']._loaded_options = None
  _globals['_CREATEAPIKEYREQUEST'].fields_by_name['parent']._serialized_options = b'\342A\001\002'
  _globals['_CREATEAPIKEYREQUEST'].fields_by_name['api_key']._loaded_options = None
  _globals['_CREATEAPIKEYREQUEST'].fields_by_name['api_key']._serialized_options = b'\342A\001\002'
  _globals['_LISTAPIKEYSREQUEST'].fields_by_name['parent']._loaded_options = None
  _globals['_LISTAPIKEYSREQUEST'].fields_by_name['parent']._serialized_options = b'\342A\001\002'
  _globals['_LISTAPIKEYSREQUEST'].fields_by_name['page_size']._loaded_options = None
  _globals['_LISTAPIKEYSREQUEST'].fields_by_name['page_size']._serialized_options = b'\342A\001\001'
  _globals['_LISTAPIKEYSREQUEST'].fields_by_name['page_token']._loaded_options = None
  _globals['_LISTAPIKEYSREQUEST'].fields_by_name['page_token']._serialized_options = b'\342A\001\001'
  _globals['_LISTAPIKEYSREQUEST'].fields_by_name['filter']._loaded_options = None
  _globals['_LISTAPIKEYSREQUEST'].fields_by_name['filter']._serialized_options = b'\342A\001\001'
  _globals['_LISTAPIKEYSREQUEST'].fields_by_name['order_by']._loaded_options = None
  _globals['_LISTAPIKEYSREQUEST'].fields_by_name['order_by']._serialized_options = b'\342A\001\001'
  _globals['_DELETEAPIKEYREQUEST'].fields_by_name['parent']._loaded_options = None
  _globals['_DELETEAPIKEYREQUEST'].fields_by_name['parent']._serialized_options = b'\342A\001\002'
  _globals['_DELETEAPIKEYREQUEST'].fields_by_name['key_id']._loaded_options = None
  _globals['_DELETEAPIKEYREQUEST'].fields_by_name['key_id']._serialized_options = b'\342A\001\002'
  _globals['_GETINTERNALAPIKEYREQUEST'].fields_by_name['parent']._loaded_options = None
  _globals['_GETINTERNALAPIKEYREQUEST'].fields_by_name['parent']._serialized_options = b'\342A\001\002'
  _globals['_UPDATEAPIKEYACCOUNTREQUEST'].fields_by_name['old_account']._loaded_options = None
  _globals['_UPDATEAPIKEYACCOUNTREQUEST'].fields_by_name['old_account']._serialized_options = b'\342A\001\002'
  _globals['_UPDATEAPIKEYACCOUNTREQUEST'].fields_by_name['new_account']._loaded_options = None
  _globals['_UPDATEAPIKEYACCOUNTREQUEST'].fields_by_name['new_account']._serialized_options = b'\342A\001\002'
  _globals['_APIKEY']._serialized_start=101
  _globals['_APIKEY']._serialized_end=282
  _globals['_CREATEAPIKEYREQUEST']._serialized_start=284
  _globals['_CREATEAPIKEYREQUEST']._serialized_end=367
  _globals['_LISTAPIKEYSREQUEST']._serialized_start=370
  _globals['_LISTAPIKEYSREQUEST']._serialized_end=509
  _globals['_LISTAPIKEYSRESPONSE']._serialized_start=511
  _globals['_LISTAPIKEYSRESPONSE']._serialized_end=612
  _globals['_DELETEAPIKEYREQUEST']._serialized_start=614
  _globals['_DELETEAPIKEYREQUEST']._serialized_end=679
  _globals['_GETINTERNALAPIKEYREQUEST']._serialized_start=681
  _globals['_GETINTERNALAPIKEYREQUEST']._serialized_end=729
  _globals['_UPDATEAPIKEYACCOUNTREQUEST']._serialized_start=731
  _globals['_UPDATEAPIKEYACCOUNTREQUEST']._serialized_end=813
# @@protoc_insertion_point(module_scope)



from . import options_pb2 as _options_pb2
from . import status_pb2 as _status_pb2
from ..google.api import field_behavior_pb2 as _field_behavior_pb2
from ..google.api import resource_pb2 as _resource_pb2
from google.protobuf import field_mask_pb2 as _field_mask_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class DeployedModel(_message.Message):
    __slots__ = ("name", "display_name", "description", "create_time", "created_by", "model", "deployment", "default", "state", "serverless", "status", "public", "update_time")
    class State(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        STATE_UNSPECIFIED: _ClassVar[DeployedModel.State]
        UNDEPLOYING: _ClassVar[DeployedModel.State]
        DEPLOYING: _ClassVar[DeployedModel.State]
        DEPLOYED: _ClassVar[DeployedModel.State]
        UPDATING: _ClassVar[DeployedModel.State]
    STATE_UNSPECIFIED: DeployedModel.State
    UNDEPLOYING: DeployedModel.State
    DEPLOYING: DeployedModel.State
    DEPLOYED: DeployedModel.State
    UPDATING: DeployedModel.State
    NAME_FIELD_NUMBER: _ClassVar[int]
    DISPLAY_NAME_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    CREATE_TIME_FIELD_NUMBER: _ClassVar[int]
    CREATED_BY_FIELD_NUMBER: _ClassVar[int]
    MODEL_FIELD_NUMBER: _ClassVar[int]
    DEPLOYMENT_FIELD_NUMBER: _ClassVar[int]
    DEFAULT_FIELD_NUMBER: _ClassVar[int]
    STATE_FIELD_NUMBER: _ClassVar[int]
    SERVERLESS_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    PUBLIC_FIELD_NUMBER: _ClassVar[int]
    UPDATE_TIME_FIELD_NUMBER: _ClassVar[int]
    name: str
    display_name: str
    description: str
    create_time: _timestamp_pb2.Timestamp
    created_by: str
    model: str
    deployment: str
    default: bool
    state: DeployedModel.State
    serverless: bool
    status: _status_pb2.Status
    public: bool
    update_time: _timestamp_pb2.Timestamp
    def __init__(self, name: _Optional[str] = ..., display_name: _Optional[str] = ..., description: _Optional[str] = ..., create_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., created_by: _Optional[str] = ..., model: _Optional[str] = ..., deployment: _Optional[str] = ..., default: bool = ..., state: _Optional[_Union[DeployedModel.State, str]] = ..., serverless: bool = ..., status: _Optional[_Union[_status_pb2.Status, _Mapping]] = ..., public: bool = ..., update_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...) -> None: ...

class DeployedModelRef(_message.Message):
    __slots__ = ("name", "deployment", "state", "default", "public")
    NAME_FIELD_NUMBER: _ClassVar[int]
    DEPLOYMENT_FIELD_NUMBER: _ClassVar[int]
    STATE_FIELD_NUMBER: _ClassVar[int]
    DEFAULT_FIELD_NUMBER: _ClassVar[int]
    PUBLIC_FIELD_NUMBER: _ClassVar[int]
    name: str
    deployment: str
    state: DeployedModel.State
    default: bool
    public: bool
    def __init__(self, name: _Optional[str] = ..., deployment: _Optional[str] = ..., state: _Optional[_Union[DeployedModel.State, str]] = ..., default: bool = ..., public: bool = ...) -> None: ...

class CreateDeployedModelRequest(_message.Message):
    __slots__ = ("parent", "deployed_model")
    PARENT_FIELD_NUMBER: _ClassVar[int]
    DEPLOYED_MODEL_FIELD_NUMBER: _ClassVar[int]
    parent: str
    deployed_model: DeployedModel
    def __init__(self, parent: _Optional[str] = ..., deployed_model: _Optional[_Union[DeployedModel, _Mapping]] = ...) -> None: ...

class DeleteDeployedModelRequest(_message.Message):
    __slots__ = ("name",)
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: str
    def __init__(self, name: _Optional[str] = ...) -> None: ...

class GetDeployedModelRequest(_message.Message):
    __slots__ = ("name",)
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: str
    def __init__(self, name: _Optional[str] = ...) -> None: ...

class ListDeployedModelsRequest(_message.Message):
    __slots__ = ("parent", "page_size", "page_token", "filter", "order_by")
    PARENT_FIELD_NUMBER: _ClassVar[int]
    PAGE_SIZE_FIELD_NUMBER: _ClassVar[int]
    PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    FILTER_FIELD_NUMBER: _ClassVar[int]
    ORDER_BY_FIELD_NUMBER: _ClassVar[int]
    parent: str
    page_size: int
    page_token: str
    filter: str
    order_by: str
    def __init__(self, parent: _Optional[str] = ..., page_size: _Optional[int] = ..., page_token: _Optional[str] = ..., filter: _Optional[str] = ..., order_by: _Optional[str] = ...) -> None: ...

class ListDeployedModelsResponse(_message.Message):
    __slots__ = ("deployed_models", "next_page_token", "total_size")
    DEPLOYED_MODELS_FIELD_NUMBER: _ClassVar[int]
    NEXT_PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    TOTAL_SIZE_FIELD_NUMBER: _ClassVar[int]
    deployed_models: _containers.RepeatedCompositeFieldContainer[DeployedModel]
    next_page_token: str
    total_size: int
    def __init__(self, deployed_models: _Optional[_Iterable[_Union[DeployedModel, _Mapping]]] = ..., next_page_token: _Optional[str] = ..., total_size: _Optional[int] = ...) -> None: ...

class UpdateDeployedModelRequest(_message.Message):
    __slots__ = ("deployed_model", "update_mask")
    DEPLOYED_MODEL_FIELD_NUMBER: _ClassVar[int]
    UPDATE_MASK_FIELD_NUMBER: _ClassVar[int]
    deployed_model: DeployedModel
    update_mask: _field_mask_pb2.FieldMask
    def __init__(self, deployed_model: _Optional[_Union[DeployedModel, _Mapping]] = ..., update_mask: _Optional[_Union[_field_mask_pb2.FieldMask, _Mapping]] = ...) -> None: ...



from ..buf.validate import validate_pb2 as _validate_pb2
from . import options_pb2 as _options_pb2
from . import status_pb2 as _status_pb2
from ..google.api import field_behavior_pb2 as _field_behavior_pb2
from ..google.api import resource_pb2 as _resource_pb2
from ..google.api import visibility_pb2 as _visibility_pb2
from google.protobuf import duration_pb2 as _duration_pb2
from google.protobuf import field_mask_pb2 as _field_mask_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Region(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    REGION_UNSPECIFIED: _ClassVar[Region]
    US_IOWA_1: _ClassVar[Region]
    US_VIRGINIA_1: _ClassVar[Region]
    US_VIRGINIA_2: _ClassVar[Region]
    US_ILLINOIS_1: _ClassVar[Region]
    AP_TOKYO_1: _ClassVar[Region]
    EU_LONDON_1: _ClassVar[Region]
    US_ARIZONA_1: _ClassVar[Region]
    US_TEXAS_1: _ClassVar[Region]
    US_ILLINOIS_2: _ClassVar[Region]
    EU_FRANKFURT_1: _ClassVar[Region]
    US_TEXAS_2: _ClassVar[Region]
    EU_PARIS_1: _ClassVar[Region]
    EU_HELSINKI_1: _ClassVar[Region]
    US_NEVADA_1: _ClassVar[Region]
    EU_ICELAND_1: _ClassVar[Region]
    EU_ICELAND_2: _ClassVar[Region]
    US_WASHINGTON_1: _ClassVar[Region]
    US_WASHINGTON_2: _ClassVar[Region]

class AcceleratorType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    ACCELERATOR_TYPE_UNSPECIFIED: _ClassVar[AcceleratorType]
    NVIDIA_A100_80GB: _ClassVar[AcceleratorType]
    NVIDIA_H100_80GB: _ClassVar[AcceleratorType]
    AMD_MI300X_192GB: _ClassVar[AcceleratorType]
    NVIDIA_A10G_24GB: _ClassVar[AcceleratorType]
    NVIDIA_A100_40GB: _ClassVar[AcceleratorType]
    NVIDIA_L4_24GB: _ClassVar[AcceleratorType]
    NVIDIA_H200_141GB: _ClassVar[AcceleratorType]
    NVIDIA_B200_180GB: _ClassVar[AcceleratorType]

class DirectRouteType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    DIRECT_ROUTE_TYPE_UNSPECIFIED: _ClassVar[DirectRouteType]
    INTERNET: _ClassVar[DirectRouteType]
    GCP_PRIVATE_SERVICE_CONNECT: _ClassVar[DirectRouteType]
    AWS_PRIVATELINK: _ClassVar[DirectRouteType]
REGION_UNSPECIFIED: Region
US_IOWA_1: Region
US_VIRGINIA_1: Region
US_VIRGINIA_2: Region
US_ILLINOIS_1: Region
AP_TOKYO_1: Region
EU_LONDON_1: Region
US_ARIZONA_1: Region
US_TEXAS_1: Region
US_ILLINOIS_2: Region
EU_FRANKFURT_1: Region
US_TEXAS_2: Region
EU_PARIS_1: Region
EU_HELSINKI_1: Region
US_NEVADA_1: Region
EU_ICELAND_1: Region
EU_ICELAND_2: Region
US_WASHINGTON_1: Region
US_WASHINGTON_2: Region
ACCELERATOR_TYPE_UNSPECIFIED: AcceleratorType
NVIDIA_A100_80GB: AcceleratorType
NVIDIA_H100_80GB: AcceleratorType
AMD_MI300X_192GB: AcceleratorType
NVIDIA_A10G_24GB: AcceleratorType
NVIDIA_A100_40GB: AcceleratorType
NVIDIA_L4_24GB: AcceleratorType
NVIDIA_H200_141GB: AcceleratorType
NVIDIA_B200_180GB: AcceleratorType
DIRECT_ROUTE_TYPE_UNSPECIFIED: DirectRouteType
INTERNET: DirectRouteType
GCP_PRIVATE_SERVICE_CONNECT: DirectRouteType
AWS_PRIVATELINK: DirectRouteType

class Deployment(_message.Message):
    __slots__ = ("name", "display_name", "description", "create_time", "expire_time", "purge_time", "delete_time", "created_by", "state", "status", "annotations", "min_replica_count", "max_replica_count", "replica_count", "autoscaling_policy", "base_model", "accelerator_count", "accelerator_type", "precision", "world_size", "generator_count", "disaggregated_prefill_count", "disaggregated_prefill_world_size", "max_batch_size", "cluster", "enable_addons", "live_merge", "draft_token_count", "draft_model", "ngram_speculation_length", "max_peft_batch_size", "kv_cache_memory_pct", "enable_session_affinity", "direct_route_api_keys", "image_tag", "num_peft_device_cached", "direct_route_type", "direct_route_handle", "deployment_template", "auto_tune", "region", "disable_accounting", "extra_args", "max_context_length", "extra_values", "engine", "update_time", "for_training")
    class State(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        STATE_UNSPECIFIED: _ClassVar[Deployment.State]
        CREATING: _ClassVar[Deployment.State]
        READY: _ClassVar[Deployment.State]
        DELETING: _ClassVar[Deployment.State]
        FAILED: _ClassVar[Deployment.State]
        UPDATING: _ClassVar[Deployment.State]
        DELETED: _ClassVar[Deployment.State]
    STATE_UNSPECIFIED: Deployment.State
    CREATING: Deployment.State
    READY: Deployment.State
    DELETING: Deployment.State
    FAILED: Deployment.State
    UPDATING: Deployment.State
    DELETED: Deployment.State
    class Precision(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        PRECISION_UNSPECIFIED: _ClassVar[Deployment.Precision]
        FP16: _ClassVar[Deployment.Precision]
        FP8: _ClassVar[Deployment.Precision]
        FP8_MM: _ClassVar[Deployment.Precision]
        FP8_AR: _ClassVar[Deployment.Precision]
        FP8_MM_KV_ATTN: _ClassVar[Deployment.Precision]
        FP8_KV: _ClassVar[Deployment.Precision]
        FP8_MM_V2: _ClassVar[Deployment.Precision]
        FP8_V2: _ClassVar[Deployment.Precision]
        FP8_MM_KV_ATTN_V2: _ClassVar[Deployment.Precision]
        NF4: _ClassVar[Deployment.Precision]
    PRECISION_UNSPECIFIED: Deployment.Precision
    FP16: Deployment.Precision
    FP8: Deployment.Precision
    FP8_MM: Deployment.Precision
    FP8_AR: Deployment.Precision
    FP8_MM_KV_ATTN: Deployment.Precision
    FP8_KV: Deployment.Precision
    FP8_MM_V2: Deployment.Precision
    FP8_V2: Deployment.Precision
    FP8_MM_KV_ATTN_V2: Deployment.Precision
    NF4: Deployment.Precision
    class Engine(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        ENGINE_UNSPECIFIED: _ClassVar[Deployment.Engine]
        FIREATTENTION: _ClassVar[Deployment.Engine]
        VLLM: _ClassVar[Deployment.Engine]
        NIM: _ClassVar[Deployment.Engine]
    ENGINE_UNSPECIFIED: Deployment.Engine
    FIREATTENTION: Deployment.Engine
    VLLM: Deployment.Engine
    NIM: Deployment.Engine
    class AnnotationsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    class ExtraValuesEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    NAME_FIELD_NUMBER: _ClassVar[int]
    DISPLAY_NAME_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    CREATE_TIME_FIELD_NUMBER: _ClassVar[int]
    EXPIRE_TIME_FIELD_NUMBER: _ClassVar[int]
    PURGE_TIME_FIELD_NUMBER: _ClassVar[int]
    DELETE_TIME_FIELD_NUMBER: _ClassVar[int]
    CREATED_BY_FIELD_NUMBER: _ClassVar[int]
    STATE_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    ANNOTATIONS_FIELD_NUMBER: _ClassVar[int]
    MIN_REPLICA_COUNT_FIELD_NUMBER: _ClassVar[int]
    MAX_REPLICA_COUNT_FIELD_NUMBER: _ClassVar[int]
    REPLICA_COUNT_FIELD_NUMBER: _ClassVar[int]
    AUTOSCALING_POLICY_FIELD_NUMBER: _ClassVar[int]
    BASE_MODEL_FIELD_NUMBER: _ClassVar[int]
    ACCELERATOR_COUNT_FIELD_NUMBER: _ClassVar[int]
    ACCELERATOR_TYPE_FIELD_NUMBER: _ClassVar[int]
    PRECISION_FIELD_NUMBER: _ClassVar[int]
    WORLD_SIZE_FIELD_NUMBER: _ClassVar[int]
    GENERATOR_COUNT_FIELD_NUMBER: _ClassVar[int]
    DISAGGREGATED_PREFILL_COUNT_FIELD_NUMBER: _ClassVar[int]
    DISAGGREGATED_PREFILL_WORLD_SIZE_FIELD_NUMBER: _ClassVar[int]
    MAX_BATCH_SIZE_FIELD_NUMBER: _ClassVar[int]
    CLUSTER_FIELD_NUMBER: _ClassVar[int]
    ENABLE_ADDONS_FIELD_NUMBER: _ClassVar[int]
    LIVE_MERGE_FIELD_NUMBER: _ClassVar[int]
    DRAFT_TOKEN_COUNT_FIELD_NUMBER: _ClassVar[int]
    DRAFT_MODEL_FIELD_NUMBER: _ClassVar[int]
    NGRAM_SPECULATION_LENGTH_FIELD_NUMBER: _ClassVar[int]
    MAX_PEFT_BATCH_SIZE_FIELD_NUMBER: _ClassVar[int]
    KV_CACHE_MEMORY_PCT_FIELD_NUMBER: _ClassVar[int]
    ENABLE_SESSION_AFFINITY_FIELD_NUMBER: _ClassVar[int]
    DIRECT_ROUTE_API_KEYS_FIELD_NUMBER: _ClassVar[int]
    IMAGE_TAG_FIELD_NUMBER: _ClassVar[int]
    NUM_PEFT_DEVICE_CACHED_FIELD_NUMBER: _ClassVar[int]
    DIRECT_ROUTE_TYPE_FIELD_NUMBER: _ClassVar[int]
    DIRECT_ROUTE_HANDLE_FIELD_NUMBER: _ClassVar[int]
    DEPLOYMENT_TEMPLATE_FIELD_NUMBER: _ClassVar[int]
    AUTO_TUNE_FIELD_NUMBER: _ClassVar[int]
    REGION_FIELD_NUMBER: _ClassVar[int]
    DISABLE_ACCOUNTING_FIELD_NUMBER: _ClassVar[int]
    EXTRA_ARGS_FIELD_NUMBER: _ClassVar[int]
    MAX_CONTEXT_LENGTH_FIELD_NUMBER: _ClassVar[int]
    EXTRA_VALUES_FIELD_NUMBER: _ClassVar[int]
    ENGINE_FIELD_NUMBER: _ClassVar[int]
    UPDATE_TIME_FIELD_NUMBER: _ClassVar[int]
    FOR_TRAINING_FIELD_NUMBER: _ClassVar[int]
    name: str
    display_name: str
    description: str
    create_time: _timestamp_pb2.Timestamp
    expire_time: _timestamp_pb2.Timestamp
    purge_time: _timestamp_pb2.Timestamp
    delete_time: _timestamp_pb2.Timestamp
    created_by: str
    state: Deployment.State
    status: _status_pb2.Status
    annotations: _containers.ScalarMap[str, str]
    min_replica_count: int
    max_replica_count: int
    replica_count: int
    autoscaling_policy: AutoscalingPolicy
    base_model: str
    accelerator_count: int
    accelerator_type: AcceleratorType
    precision: Deployment.Precision
    world_size: int
    generator_count: int
    disaggregated_prefill_count: int
    disaggregated_prefill_world_size: int
    max_batch_size: int
    cluster: str
    enable_addons: bool
    live_merge: bool
    draft_token_count: int
    draft_model: str
    ngram_speculation_length: int
    max_peft_batch_size: int
    kv_cache_memory_pct: int
    enable_session_affinity: bool
    direct_route_api_keys: _containers.RepeatedScalarFieldContainer[str]
    image_tag: str
    num_peft_device_cached: int
    direct_route_type: DirectRouteType
    direct_route_handle: str
    deployment_template: str
    auto_tune: AutoTune
    region: Region
    disable_accounting: bool
    extra_args: _containers.RepeatedScalarFieldContainer[str]
    max_context_length: int
    extra_values: _containers.ScalarMap[str, str]
    engine: Deployment.Engine
    update_time: _timestamp_pb2.Timestamp
    for_training: bool
    def __init__(self, name: _Optional[str] = ..., display_name: _Optional[str] = ..., description: _Optional[str] = ..., create_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., expire_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., purge_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., delete_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., created_by: _Optional[str] = ..., state: _Optional[_Union[Deployment.State, str]] = ..., status: _Optional[_Union[_status_pb2.Status, _Mapping]] = ..., annotations: _Optional[_Mapping[str, str]] = ..., min_replica_count: _Optional[int] = ..., max_replica_count: _Optional[int] = ..., replica_count: _Optional[int] = ..., autoscaling_policy: _Optional[_Union[AutoscalingPolicy, _Mapping]] = ..., base_model: _Optional[str] = ..., accelerator_count: _Optional[int] = ..., accelerator_type: _Optional[_Union[AcceleratorType, str]] = ..., precision: _Optional[_Union[Deployment.Precision, str]] = ..., world_size: _Optional[int] = ..., generator_count: _Optional[int] = ..., disaggregated_prefill_count: _Optional[int] = ..., disaggregated_prefill_world_size: _Optional[int] = ..., max_batch_size: _Optional[int] = ..., cluster: _Optional[str] = ..., enable_addons: bool = ..., live_merge: bool = ..., draft_token_count: _Optional[int] = ..., draft_model: _Optional[str] = ..., ngram_speculation_length: _Optional[int] = ..., max_peft_batch_size: _Optional[int] = ..., kv_cache_memory_pct: _Optional[int] = ..., enable_session_affinity: bool = ..., direct_route_api_keys: _Optional[_Iterable[str]] = ..., image_tag: _Optional[str] = ..., num_peft_device_cached: _Optional[int] = ..., direct_route_type: _Optional[_Union[DirectRouteType, str]] = ..., direct_route_handle: _Optional[str] = ..., deployment_template: _Optional[str] = ..., auto_tune: _Optional[_Union[AutoTune, _Mapping]] = ..., region: _Optional[_Union[Region, str]] = ..., disable_accounting: bool = ..., extra_args: _Optional[_Iterable[str]] = ..., max_context_length: _Optional[int] = ..., extra_values: _Optional[_Mapping[str, str]] = ..., engine: _Optional[_Union[Deployment.Engine, str]] = ..., update_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., for_training: bool = ...) -> None: ...

class AutoTune(_message.Message):
    __slots__ = ("long_prompt",)
    LONG_PROMPT_FIELD_NUMBER: _ClassVar[int]
    long_prompt: bool
    def __init__(self, long_prompt: bool = ...) -> None: ...

class AutoscalingPolicy(_message.Message):
    __slots__ = ("scale_up_window", "scale_down_window", "scale_to_zero_window", "load_targets")
    class LoadTargetsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: float
        def __init__(self, key: _Optional[str] = ..., value: _Optional[float] = ...) -> None: ...
    SCALE_UP_WINDOW_FIELD_NUMBER: _ClassVar[int]
    SCALE_DOWN_WINDOW_FIELD_NUMBER: _ClassVar[int]
    SCALE_TO_ZERO_WINDOW_FIELD_NUMBER: _ClassVar[int]
    LOAD_TARGETS_FIELD_NUMBER: _ClassVar[int]
    scale_up_window: _duration_pb2.Duration
    scale_down_window: _duration_pb2.Duration
    scale_to_zero_window: _duration_pb2.Duration
    load_targets: _containers.ScalarMap[str, float]
    def __init__(self, scale_up_window: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., scale_down_window: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., scale_to_zero_window: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., load_targets: _Optional[_Mapping[str, float]] = ...) -> None: ...

class CreateDeploymentRequest(_message.Message):
    __slots__ = ("parent", "deployment", "disable_auto_deploy", "disable_speculative_decoding", "deployment_id")
    PARENT_FIELD_NUMBER: _ClassVar[int]
    DEPLOYMENT_FIELD_NUMBER: _ClassVar[int]
    DISABLE_AUTO_DEPLOY_FIELD_NUMBER: _ClassVar[int]
    DISABLE_SPECULATIVE_DECODING_FIELD_NUMBER: _ClassVar[int]
    DEPLOYMENT_ID_FIELD_NUMBER: _ClassVar[int]
    parent: str
    deployment: Deployment
    disable_auto_deploy: bool
    disable_speculative_decoding: bool
    deployment_id: str
    def __init__(self, parent: _Optional[str] = ..., deployment: _Optional[_Union[Deployment, _Mapping]] = ..., disable_auto_deploy: bool = ..., disable_speculative_decoding: bool = ..., deployment_id: _Optional[str] = ...) -> None: ...

class GetDeploymentRequest(_message.Message):
    __slots__ = ("name",)
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: str
    def __init__(self, name: _Optional[str] = ...) -> None: ...

class ListDeploymentsRequest(_message.Message):
    __slots__ = ("parent", "page_size", "page_token", "filter", "order_by", "show_deleted", "show_internal")
    PARENT_FIELD_NUMBER: _ClassVar[int]
    PAGE_SIZE_FIELD_NUMBER: _ClassVar[int]
    PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    FILTER_FIELD_NUMBER: _ClassVar[int]
    ORDER_BY_FIELD_NUMBER: _ClassVar[int]
    SHOW_DELETED_FIELD_NUMBER: _ClassVar[int]
    SHOW_INTERNAL_FIELD_NUMBER: _ClassVar[int]
    parent: str
    page_size: int
    page_token: str
    filter: str
    order_by: str
    show_deleted: bool
    show_internal: bool
    def __init__(self, parent: _Optional[str] = ..., page_size: _Optional[int] = ..., page_token: _Optional[str] = ..., filter: _Optional[str] = ..., order_by: _Optional[str] = ..., show_deleted: bool = ..., show_internal: bool = ...) -> None: ...

class ListDeploymentsResponse(_message.Message):
    __slots__ = ("deployments", "next_page_token", "total_size")
    DEPLOYMENTS_FIELD_NUMBER: _ClassVar[int]
    NEXT_PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    TOTAL_SIZE_FIELD_NUMBER: _ClassVar[int]
    deployments: _containers.RepeatedCompositeFieldContainer[Deployment]
    next_page_token: str
    total_size: int
    def __init__(self, deployments: _Optional[_Iterable[_Union[Deployment, _Mapping]]] = ..., next_page_token: _Optional[str] = ..., total_size: _Optional[int] = ...) -> None: ...

class UpdateDeploymentRequest(_message.Message):
    __slots__ = ("deployment", "update_mask")
    DEPLOYMENT_FIELD_NUMBER: _ClassVar[int]
    UPDATE_MASK_FIELD_NUMBER: _ClassVar[int]
    deployment: Deployment
    update_mask: _field_mask_pb2.FieldMask
    def __init__(self, deployment: _Optional[_Union[Deployment, _Mapping]] = ..., update_mask: _Optional[_Union[_field_mask_pb2.FieldMask, _Mapping]] = ...) -> None: ...

class DeleteDeploymentRequest(_message.Message):
    __slots__ = ("name", "hard", "ignore_checks")
    NAME_FIELD_NUMBER: _ClassVar[int]
    HARD_FIELD_NUMBER: _ClassVar[int]
    IGNORE_CHECKS_FIELD_NUMBER: _ClassVar[int]
    name: str
    hard: bool
    ignore_checks: bool
    def __init__(self, name: _Optional[str] = ..., hard: bool = ..., ignore_checks: bool = ...) -> None: ...

class UndeleteDeploymentRequest(_message.Message):
    __slots__ = ("name",)
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: str
    def __init__(self, name: _Optional[str] = ...) -> None: ...

class ScaleDeploymentRequest(_message.Message):
    __slots__ = ("name", "replica_count")
    NAME_FIELD_NUMBER: _ClassVar[int]
    REPLICA_COUNT_FIELD_NUMBER: _ClassVar[int]
    name: str
    replica_count: int
    def __init__(self, name: _Optional[str] = ..., replica_count: _Optional[int] = ...) -> None: ...

class GetDeploymentMetricsRequest(_message.Message):
    __slots__ = ("name", "time_range")
    NAME_FIELD_NUMBER: _ClassVar[int]
    TIME_RANGE_FIELD_NUMBER: _ClassVar[int]
    name: str
    time_range: str
    def __init__(self, name: _Optional[str] = ..., time_range: _Optional[str] = ...) -> None: ...

class GetDeploymentMetricsResponse(_message.Message):
    __slots__ = ("metrics",)
    class MetricsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: float
        def __init__(self, key: _Optional[str] = ..., value: _Optional[float] = ...) -> None: ...
    METRICS_FIELD_NUMBER: _ClassVar[int]
    metrics: _containers.ScalarMap[str, float]
    def __init__(self, metrics: _Optional[_Mapping[str, float]] = ...) -> None: ...

class GetDeploymentPrerequisitesRequest(_message.Message):
    __slots__ = ("parent", "name")
    PARENT_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    parent: str
    name: str
    def __init__(self, parent: _Optional[str] = ..., name: _Optional[str] = ...) -> None: ...

class DeploymentPrerequisites(_message.Message):
    __slots__ = ("accelerator_configs",)
    ACCELERATOR_CONFIGS_FIELD_NUMBER: _ClassVar[int]
    accelerator_configs: _containers.RepeatedCompositeFieldContainer[DeploymentAcceleratorConfig]
    def __init__(self, accelerator_configs: _Optional[_Iterable[_Union[DeploymentAcceleratorConfig, _Mapping]]] = ...) -> None: ...

class DeploymentAcceleratorConfig(_message.Message):
    __slots__ = ("accelerator_type", "precision", "min_accelerator_count", "regions")
    ACCELERATOR_TYPE_FIELD_NUMBER: _ClassVar[int]
    PRECISION_FIELD_NUMBER: _ClassVar[int]
    MIN_ACCELERATOR_COUNT_FIELD_NUMBER: _ClassVar[int]
    REGIONS_FIELD_NUMBER: _ClassVar[int]
    accelerator_type: AcceleratorType
    precision: Deployment.Precision
    min_accelerator_count: int
    regions: _containers.RepeatedScalarFieldContainer[Region]
    def __init__(self, accelerator_type: _Optional[_Union[AcceleratorType, str]] = ..., precision: _Optional[_Union[Deployment.Precision, str]] = ..., min_accelerator_count: _Optional[int] = ..., regions: _Optional[_Iterable[_Union[Region, str]]] = ...) -> None: ...



from . import deployment_pb2 as _deployment_pb2
from . import options_pb2 as _options_pb2
from ..google.api import field_behavior_pb2 as _field_behavior_pb2
from ..google.api import resource_pb2 as _resource_pb2
from ..google.api import visibility_pb2 as _visibility_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Reservation(_message.Message):
    __slots__ = ("name", "display_name", "description", "create_time", "created_by", "region", "accelerator_type", "reserved_count", "start_time", "end_time", "contract", "update_time")
    NAME_FIELD_NUMBER: _ClassVar[int]
    DISPLAY_NAME_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    CREATE_TIME_FIELD_NUMBER: _ClassVar[int]
    CREATED_BY_FIELD_NUMBER: _ClassVar[int]
    REGION_FIELD_NUMBER: _ClassVar[int]
    ACCELERATOR_TYPE_FIELD_NUMBER: _ClassVar[int]
    RESERVED_COUNT_FIELD_NUMBER: _ClassVar[int]
    START_TIME_FIELD_NUMBER: _ClassVar[int]
    END_TIME_FIELD_NUMBER: _ClassVar[int]
    CONTRACT_FIELD_NUMBER: _ClassVar[int]
    UPDATE_TIME_FIELD_NUMBER: _ClassVar[int]
    name: str
    display_name: str
    description: str
    create_time: _timestamp_pb2.Timestamp
    created_by: str
    region: _deployment_pb2.Region
    accelerator_type: _deployment_pb2.AcceleratorType
    reserved_count: int
    start_time: _timestamp_pb2.Timestamp
    end_time: _timestamp_pb2.Timestamp
    contract: str
    update_time: _timestamp_pb2.Timestamp
    def __init__(self, name: _Optional[str] = ..., display_name: _Optional[str] = ..., description: _Optional[str] = ..., create_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., created_by: _Optional[str] = ..., region: _Optional[_Union[_deployment_pb2.Region, str]] = ..., accelerator_type: _Optional[_Union[_deployment_pb2.AcceleratorType, str]] = ..., reserved_count: _Optional[int] = ..., start_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., end_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., contract: _Optional[str] = ..., update_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...) -> None: ...

class GetReservationRequest(_message.Message):
    __slots__ = ("name",)
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: str
    def __init__(self, name: _Optional[str] = ...) -> None: ...

class ListReservationsRequest(_message.Message):
    __slots__ = ("parent", "page_size", "page_token", "filter", "order_by")
    PARENT_FIELD_NUMBER: _ClassVar[int]
    PAGE_SIZE_FIELD_NUMBER: _ClassVar[int]
    PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    FILTER_FIELD_NUMBER: _ClassVar[int]
    ORDER_BY_FIELD_NUMBER: _ClassVar[int]
    parent: str
    page_size: int
    page_token: str
    filter: str
    order_by: str
    def __init__(self, parent: _Optional[str] = ..., page_size: _Optional[int] = ..., page_token: _Optional[str] = ..., filter: _Optional[str] = ..., order_by: _Optional[str] = ...) -> None: ...

class ListReservationsResponse(_message.Message):
    __slots__ = ("reservations", "next_page_token", "total_size")
    RESERVATIONS_FIELD_NUMBER: _ClassVar[int]
    NEXT_PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    TOTAL_SIZE_FIELD_NUMBER: _ClassVar[int]
    reservations: _containers.RepeatedCompositeFieldContainer[Reservation]
    next_page_token: str
    total_size: int
    def __init__(self, reservations: _Optional[_Iterable[_Union[Reservation, _Mapping]]] = ..., next_page_token: _Optional[str] = ..., total_size: _Optional[int] = ...) -> None: ...



from ..google.api import field_behavior_pb2 as _field_behavior_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class WandbConfig(_message.Message):
    __slots__ = ("enabled", "api_key", "project", "entity", "run_id", "url")
    ENABLED_FIELD_NUMBER: _ClassVar[int]
    API_KEY_FIELD_NUMBER: _ClassVar[int]
    PROJECT_FIELD_NUMBER: _ClassVar[int]
    ENTITY_FIELD_NUMBER: _ClassVar[int]
    RUN_ID_FIELD_NUMBER: _ClassVar[int]
    URL_FIELD_NUMBER: _ClassVar[int]
    enabled: bool
    api_key: str
    project: str
    entity: str
    run_id: str
    url: str
    def __init__(self, enabled: bool = ..., api_key: _Optional[str] = ..., project: _Optional[str] = ..., entity: _Optional[str] = ..., run_id: _Optional[str] = ..., url: _Optional[str] = ...) -> None: ...



from ..google.api import field_behavior_pb2 as _field_behavior_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Secret(_message.Message):
    __slots__ = ("name", "key_name", "value")
    NAME_FIELD_NUMBER: _ClassVar[int]
    KEY_NAME_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    name: str
    key_name: str
    value: str
    def __init__(self, name: _Optional[str] = ..., key_name: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

class CreateSecretRequest(_message.Message):
    __slots__ = ("parent", "secret")
    PARENT_FIELD_NUMBER: _ClassVar[int]
    SECRET_FIELD_NUMBER: _ClassVar[int]
    parent: str
    secret: Secret
    def __init__(self, parent: _Optional[str] = ..., secret: _Optional[_Union[Secret, _Mapping]] = ...) -> None: ...

class GetSecretRequest(_message.Message):
    __slots__ = ("name",)
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: str
    def __init__(self, name: _Optional[str] = ...) -> None: ...

class ListSecretsRequest(_message.Message):
    __slots__ = ("parent", "page_size", "page_token", "filter", "order_by")
    PARENT_FIELD_NUMBER: _ClassVar[int]
    PAGE_SIZE_FIELD_NUMBER: _ClassVar[int]
    PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    FILTER_FIELD_NUMBER: _ClassVar[int]
    ORDER_BY_FIELD_NUMBER: _ClassVar[int]
    parent: str
    page_size: int
    page_token: str
    filter: str
    order_by: str
    def __init__(self, parent: _Optional[str] = ..., page_size: _Optional[int] = ..., page_token: _Optional[str] = ..., filter: _Optional[str] = ..., order_by: _Optional[str] = ...) -> None: ...

class ListSecretsResponse(_message.Message):
    __slots__ = ("secrets", "next_page_token", "total_size")
    SECRETS_FIELD_NUMBER: _ClassVar[int]
    NEXT_PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    TOTAL_SIZE_FIELD_NUMBER: _ClassVar[int]
    secrets: _containers.RepeatedCompositeFieldContainer[Secret]
    next_page_token: str
    total_size: int
    def __init__(self, secrets: _Optional[_Iterable[_Union[Secret, _Mapping]]] = ..., next_page_token: _Optional[str] = ..., total_size: _Optional[int] = ...) -> None: ...

class UpdateSecretRequest(_message.Message):
    __slots__ = ("secret",)
    SECRET_FIELD_NUMBER: _ClassVar[int]
    secret: Secret
    def __init__(self, secret: _Optional[_Union[Secret, _Mapping]] = ...) -> None: ...

class DeleteSecretRequest(_message.Message):
    __slots__ = ("name",)
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: str
    def __init__(self, name: _Optional[str] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: gateway/test_resource.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'gateway/test_resource.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import options_pb2 as gateway_dot_options__pb2
from ..google.api import resource_pb2 as google_dot_api_dot_resource__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from ..google.type import date_pb2 as google_dot_type_dot_date__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1bgateway/test_resource.proto\x12\x07gateway\x1a\x15gateway/options.proto\x1a\x19google/api/resource.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x16google/type/date.proto\"\x8c\x05\n\x0cTestResource\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\n\n\x02\x66\x32\x18\x02 \x01(\x01\x12\n\n\x02\x66\x33\x18\x03 \x01(\x02\x12\n\n\x02\x66\x34\x18\x04 \x01(\x05\x12\n\n\x02\x66\x35\x18\x05 \x01(\x03\x12\n\n\x02\x66\x36\x18\x06 \x01(\r\x12\n\n\x02\x66\x37\x18\x07 \x01(\x04\x12\n\n\x02\x66\x38\x18\x08 \x01(\x11\x12\n\n\x02\x66\x39\x18\t \x01(\x12\x12\x0b\n\x03\x66\x31\x30\x18\n \x01(\x07\x12\x0b\n\x03\x66\x31\x31\x18\x0b \x01(\x06\x12\x0b\n\x03\x66\x31\x32\x18\x0c \x01(\x0f\x12\x0b\n\x03\x66\x31\x33\x18\r \x01(\x10\x12\x0b\n\x03\x66\x31\x34\x18\x0e \x01(\x08\x12\x0b\n\x03\x66\x31\x35\x18\x0f \x01(\t\x12\'\n\x03\x66\x31\x36\x18\x10 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x03\x66\x31\x38\x18\x12 \x01(\x0e\x32\x1f.gateway.TestResource.InnerEnum\x12\x1f\n\x03\x66\x31\x39\x18\x13 \x01(\x0e\x32\x12.gateway.OuterEnum\x12/\n\x03\x66\x32\x30\x18\x14 \x01(\x0b\x32\".gateway.TestResource.InnerMessage\x12\"\n\x03\x66\x32\x31\x18\x15 \x01(\x0b\x32\x15.gateway.OuterMessage\x12\x1e\n\x03\x66\x32\x32\x18\x16 \x01(\x0b\x32\x11.google.type.Date\x1a\x1b\n\x0cInnerMessage\x12\x0b\n\x03\x66oo\x18\x01 \x01(\t\"0\n\tInnerEnum\x12\x1a\n\x16INNER_ENUM_UNSPECIFIED\x10\x00\x12\x07\n\x03ONE\x10\x01:\x83\x01\xea\x41q\n\x1d\x61pi.fireworks.ai/TestResource\x12\x33\x61\x63\x63ounts/{AccountId}/testResources/{TestResourceId}*\rtestResources2\x0ctestResource\x82\xf1\x04\x0b\n\x07\x41\x63\x63ount\x18\x01\"\x1b\n\x0cOuterMessage\x12\x0b\n\x03\x62\x61r\x18\x01 \x01(\t*0\n\tOuterEnum\x12\x1a\n\x16OUTER_ENUM_UNSPECIFIED\x10\x00\x12\x07\n\x03ONE\x10\x01\x42\x43ZAgithub.com/fw-ai/fireworks/control_plane/protos/generated/gatewayb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'gateway.test_resource_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'ZAgithub.com/fw-ai/fireworks/control_plane/protos/generated/gateway'
  _globals['_TESTRESOURCE']._loaded_options = None
  _globals['_TESTRESOURCE']._serialized_options = b'\352Aq\n\035api.fireworks.ai/TestResource\0223accounts/{AccountId}/testResources/{TestResourceId}*\rtestResources2\014testResource\202\361\004\013\n\007Account\030\001'
  _globals['_OUTERENUM']._serialized_start=831
  _globals['_OUTERENUM']._serialized_end=879
  _globals['_TESTRESOURCE']._serialized_start=148
  _globals['_TESTRESOURCE']._serialized_end=800
  _globals['_TESTRESOURCE_INNERMESSAGE']._serialized_start=589
  _globals['_TESTRESOURCE_INNERMESSAGE']._serialized_end=616
  _globals['_TESTRESOURCE_INNERENUM']._serialized_start=618
  _globals['_TESTRESOURCE_INNERENUM']._serialized_end=666
  _globals['_OUTERMESSAGE']._serialized_start=802
  _globals['_OUTERMESSAGE']._serialized_end=829
# @@protoc_insertion_point(module_scope)



from collections import defaultdict, deque
from datetime import timedelta
import inspect
import time

from fireworks._async import asyncf
from ._list_fireworks_models_response_cached import models
from fireworks.client.api_client import FireworksClient
from fireworks.client.chat import Chat as FireworksChat
from fireworks.client.chat_completion import ChatCompletionV2 as FireworksChatCompletion
from asyncstdlib.functools import cache
from functools import cache as sync_cache
from google.protobuf.field_mask_pb2 import FieldMask as SyncFieldMask
from typing import (
    AsyncGenerator,
    Generator,
    Iterable,
    List,
    Literal,
    Optional,
    Union,
    overload,
)
from fireworks.client.error import BadGatewayError, ServiceUnavailableError
from fireworks.dataset import Dataset
from fireworks.supervised_fine_tuning_job import SupervisedFineTuningJob
from fireworks.gateway import Gateway
from google.protobuf.duration_pb2 import Duration
from fireworks.control_plane.generated.protos_grpcio.gateway.deployment_pb2 import (
    ListDeploymentsRequest as SyncListDeploymentsRequest,
    UpdateDeploymentRequest as SyncUpdateDeploymentRequest,
    Deployment as SyncDeployment,
    AutoscalingPolicy as SyncAutoscalingPolicy,
    AcceleratorType as SyncAcceleratorType,
    Region as SyncRegion,
    DirectRouteType as SyncDirectRouteType,
    AutoTune as SyncAutoTune,
)
from fireworks.control_plane.generated.protos.gateway import (
    AcceleratorType as AcceleratorTypeEnum,
    AutoTune,
    AutoscalingPolicy,
    CreateSupervisedFineTuningJobRequest,
    DeployedModelState,
    Deployment,
    DeploymentPrecision,
    DeploymentState,
    DirectRouteType,
    JobState,
    ListSupervisedFineTuningJobsRequest,
    Model,
    Region,
    SupervisedFineTuningJobWeightPrecision,
    WandbConfig,
)
from fireworks.supervised_fine_tuning_job.SupervisedFineTuningJob import SupervisedFineTuningJobWeightPrecisionLiteral
import asyncio
import logging
import os
import atexit
from openai.types.chat.chat_completion import ChatCompletion as OpenAIChatCompletion
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk
from openai.types.chat.chat_completion_message_param import ChatCompletionMessageParam
from openai.types.chat.completion_create_params import ResponseFormat
from openai import NOT_GIVEN, NotGiven
from fireworks._util import async_to_sync
import sysconfig
from fireworks._literals import RegionLiteral, AcceleratorTypeLiteral, ReasoningEffort

# Configure logger with a consistent format for better debugging
logger = logging.getLogger(__name__)
handler = logging.StreamHandler()
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
handler.setFormatter(formatter)
logger.addHandler(handler)
logger.propagate = False  # Prevent duplicate logs

if os.environ.get("FIREWORKS_SDK_DEBUG"):
    logger.setLevel(logging.DEBUG)


class ChatCompletion:
    def __init__(self, llm: "LLM"):
        self._client = FireworksChatCompletion(llm._client)
        self._llm = llm

    @overload
    def create(
        self,
        messages: Iterable[ChatCompletionMessageParam],
        stream: Literal[False] = False,
        response_format: Optional[ResponseFormat] = None,
        reasoning_effort: Optional[ReasoningEffort] = None,
        max_tokens: Optional[int] = None,
        extra_headers=None,
        **kwargs,
    ) -> OpenAIChatCompletion:
        pass

    @overload
    def create(
        self,
        messages: Iterable[ChatCompletionMessageParam],
        stream: Literal[True] = True,
        response_format: Optional[ResponseFormat] = None,
        reasoning_effort: Optional[ReasoningEffort] = None,
        max_tokens: Optional[int] = None,
        extra_headers=None,
        **kwargs,
    ) -> Generator[ChatCompletionChunk, None, None]:
        pass

    def create(
        self,
        messages: Iterable[ChatCompletionMessageParam],
        stream: bool = False,
        response_format: Optional[ResponseFormat] = None,
        reasoning_effort: Optional[ReasoningEffort] = None,
        max_tokens: Optional[int] = None,
        extra_headers=None,
        **kwargs,
    ) -> Union[OpenAIChatCompletion, Generator[ChatCompletionChunk, None, None]]:
        model_id = self._create_setup()
        retries = 0
        while retries < self._llm.max_retries:
            try:
                if self._llm.enable_metrics:
                    start_time = time.time()
                result = self._client.create(
                    model=model_id,
                    prompt_or_messages=messages,
                    stream=stream,
                    extra_headers=extra_headers,
                    response_format=response_format,
                    reasoning_effort=reasoning_effort,
                    max_tokens=max_tokens,
                    **kwargs,
                )
                if self._llm.enable_metrics:
                    end_time = time.time()
                    self._llm._metrics.add_metric("time_to_last_token", end_time - start_time)
                return result
            except BadGatewayError:
                retries += 1
            except ServiceUnavailableError:
                retries += 1
        raise Exception(f"Failed to create chat completion after {self._llm.max_retries} retries")

    def _create_setup(self):
        """
        Wraps all the async code in one function to make it easier to call
        """
        self._llm._ensure_deployment_ready()
        model_id = self._llm.model_id()
        return model_id

    @overload
    async def acreate(
        self,
        messages: Iterable[ChatCompletionMessageParam],
        stream: Literal[False] = False,
        response_format: Optional[ResponseFormat] = None,
        reasoning_effort: Optional[ReasoningEffort] = None,
        max_tokens: Optional[int] = None,
        extra_headers=None,
        **kwargs,
    ) -> OpenAIChatCompletion: ...

    @overload
    async def acreate(
        self,
        messages: Iterable[ChatCompletionMessageParam],
        stream: Literal[True] = True,
        response_format: Optional[ResponseFormat] = None,
        reasoning_effort: Optional[ReasoningEffort] = None,
        max_tokens: Optional[int] = None,
        extra_headers=None,
        **kwargs,
    ) -> AsyncGenerator[ChatCompletionChunk, None]:
        pass

    async def acreate(
        self,
        messages: Iterable[ChatCompletionMessageParam],
        stream: bool = False,
        response_format: Optional[ResponseFormat] = None,
        reasoning_effort: Optional[ReasoningEffort] = None,
        max_tokens: Optional[int] = None,
        extra_headers=None,
        **kwargs,
    ) -> Union[OpenAIChatCompletion, AsyncGenerator[ChatCompletionChunk, None]]:
        self._llm._ensure_deployment_ready()
        model_id = self._llm.model_id()
        retries = 0
        while retries < self._llm.max_retries:
            try:
                resp_or_generator = self._client.acreate(  # type: ignore
                    model=model_id,
                    prompt_or_messages=messages,
                    stream=stream,
                    extra_headers=extra_headers,
                    response_format=response_format,
                    reasoning_effort=reasoning_effort,
                    max_tokens=max_tokens,
                    **kwargs,
                )
                if stream:
                    return resp_or_generator  # type: ignore
                else:
                    if self._llm.enable_metrics:
                        start_time = time.time()
                    resp = await resp_or_generator  # type: ignore
                    if self._llm.enable_metrics:
                        end_time = time.time()
                        self._llm._metrics.add_metric("time_to_last_token", end_time - start_time)
                    return resp
            except BadGatewayError as e:
                logger.debug(f"BadGatewayError: {e}. model_id: {model_id}")
                retries += 1
            except ServiceUnavailableError as e:
                logger.debug(f"ServiceUnavailableError: {e}. model_id: {model_id}")
                retries += 1
        raise Exception(f"Failed to create chat completion after {self._llm.max_retries} retries")


class Chat:
    def __init__(self, llm: "LLM", model):
        self.completions = ChatCompletion(llm)


# Make param a literal type so developer does not need to import AcceleratorType enum


DEFAULT_MAX_RETRIES = 3


class Metrics:
    """
    A class for tracking and analyzing performance metrics for LLM operations.

    This class maintains a rolling window of metrics such as response times,
    token generation speeds, and other performance indicators. It provides
    statistical methods to analyze these metrics (mean, median, min, max).

    Each metric is stored as a list of values with a maximum size to prevent
    unbounded memory growth. When the maximum size is reached, the oldest
    values are removed in a FIFO manner.
    """

    def __init__(self, max_metrics: int = 1000):
        """
        Initialize the metrics tracker.

        Args:
            max_metrics: Maximum number of values to store for each metric.
                         Defaults to 1000. When exceeded, oldest values are removed.
        """
        self._metrics = defaultdict(deque)
        self._max_metrics = max_metrics

    def add_metric(self, metric_name: str, metric_value: float):
        """
        Add a new metric value to the specified metric.

        Args:
            metric_name: Name of the metric to track
            metric_value: Value to add to the metric
        """
        self._metrics[metric_name].append(metric_value)
        if len(self._metrics[metric_name]) > self._max_metrics:
            self._metrics[metric_name].popleft()

    def get_metric(self, metric_name: str):
        """
        Get all recorded values for a specific metric.

        Args:
            metric_name: Name of the metric to retrieve

        Returns:
            List of metric values or None if the metric doesn't exist
        """
        if metric_name not in self._metrics:
            return None
        return self._metrics[metric_name]

    def get_metric_mean(self, metric_name: str) -> Optional[float]:
        """
        Calculate the arithmetic mean of a metric's values.

        Args:
            metric_name: Name of the metric

        Returns:
            Mean value or None if the metric doesn't exist
        """
        if metric_name not in self._metrics:
            return None
        return sum(self._metrics[metric_name]) / len(self._metrics[metric_name])

    def get_metric_median(self, metric_name: str):
        """
        Calculate the median value of a metric.

        Args:
            metric_name: Name of the metric

        Returns:
            Median value or None if the metric doesn't exist
        """
        if metric_name not in self._metrics:
            return None
        return sorted(self._metrics[metric_name])[len(self._metrics[metric_name]) // 2]

    def get_metric_min(self, metric_name: str):
        """
        Get the minimum value recorded for a metric.

        Args:
            metric_name: Name of the metric

        Returns:
            Minimum value or None if the metric doesn't exist
        """
        if metric_name not in self._metrics:
            return None
        return min(self._metrics[metric_name])

    def get_metric_max(self, metric_name: str):
        """
        Get the maximum value recorded for a metric.

        Args:
            metric_name: Name of the metric

        Returns:
            Maximum value or None if the metric doesn't exist
        """
        if metric_name not in self._metrics:
            return None
        return max(self._metrics[metric_name])


class DeploymentWithIdAndName(Deployment):
    def __init__(self, deployment: SyncDeployment, gateway: Gateway):
        self.id = deployment.name
        self.name = deployment.display_name
        self._gateway = gateway


class LLM:
    def __init__(
        self,
        model: str,
        name: Optional[str] = None,
        api_key: Optional[str] = None,
        base_url: str = "https://api.fireworks.ai/inference/v1",
        deployment_type: Literal["serverless", "on-demand"] = "serverless",
        accelerator_type: Union[AcceleratorTypeLiteral, NotGiven] = NOT_GIVEN,
        scale_up_window: timedelta = timedelta(seconds=1),
        scale_down_window: timedelta = timedelta(minutes=1),
        scale_to_zero_window: timedelta = timedelta(minutes=5),
        enable_metrics: bool = False,
        max_retries: int = DEFAULT_MAX_RETRIES,
        region: Optional[RegionLiteral] = None,
        description: Optional[str] = None,
        annotations: Optional[dict[str, str]] = None,
        min_replica_count: Optional[int] = None,
        max_replica_count: Optional[int] = None,
        replica_count: Optional[int] = None,
        accelerator_count: Optional[int] = None,
        precision: Optional[
            Literal[
                "FP16",
                "FP8",
                "FP8_MM",
                "FP8_AR",
                "FP8_MM_KV_ATTN",
                "FP8_KV",
                "FP8_MM_V2",
                "FP8_V2",
                "FP8_MM_KV_ATTN_V2",
                "NF4",
            ]
        ] = None,
        world_size: Optional[int] = None,
        generator_count: Optional[int] = None,
        disaggregated_prefill_count: Optional[int] = None,
        disaggregated_prefill_world_size: Optional[int] = None,
        max_batch_size: Optional[int] = None,
        cluster: Optional[str] = None,
        enable_addons: Optional[bool] = None,
        live_merge: Optional[bool] = None,
        draft_token_count: Optional[int] = None,
        draft_model: Optional[str] = None,
        ngram_speculation_length: Optional[int] = None,
        max_peft_batch_size: Optional[int] = None,
        kv_cache_memory_pct: Optional[int] = None,
        enable_session_affinity: Optional[bool] = None,
        direct_route_api_keys: Optional[list[str]] = None,
        num_peft_device_cached: Optional[int] = None,
        direct_route_type: Optional[Literal["INTERNET", "GCP_PRIVATE_SERVICE_CONNECT", "AWS_PRIVATELINK"]] = None,
        direct_route_handle: Optional[str] = None,
        long_prompt_optimized: Optional[bool] = None,
    ):
        """
        Initialize the LLM.

        Args:
            model: The model to use.
            api_key: The API key to use.
            base_url: The base URL to use.
            deployment_type: The deployment type to use.
            accelerator_type: The accelerator type to use.
            scale_up_window: The scale up window to use.
            scale_down_window: The scale down window to use.
            scale_to_zero_window: The scale to zero window to use.
            enable_metrics: Whether to enable metrics. Only records time to last token for non-streaming requests.
                TODO: add support for streaming requests.
                TODO: add support for more metrics (TTFT, Tokens/second)
            max_retries: The maximum number of retries to use.
        """
        if not model:
            raise ValueError("model is required")
        self._client = FireworksClient(api_key=api_key, base_url=base_url)
        if name is not None and name == "":
            raise ValueError("name must be non-empty")
        self._name = name
        self._model = model
        self.chat = Chat(self, self.model)
        self.deployment_type = deployment_type
        self.max_retries = max_retries
        self.enable_metrics = enable_metrics
        self._gateway = Gateway(api_key=api_key)
        self._metrics = Metrics()

        # This needs to be run in __init__ to ensure we capture deployment name
        # inside of this thread
        self._get_deployment_name()

        # Add a lock to prevent concurrent calls to _ensure_deployment_ready
        self._deployment_lock = asyncio.Lock()
        atexit.register(self._gateway._channel.close)

        if isinstance(accelerator_type, str):
            self._accelerator_type = AcceleratorTypeEnum.from_string(accelerator_type)
            self._validate_model_for_gpu(self.model, self._accelerator_type)
            self._accelerator_type_sync: SyncAcceleratorType = getattr(SyncAcceleratorType, accelerator_type)
        else:
            self._accelerator_type = accelerator_type

        # aggressive defaults for experimentation to save on cost
        self._autoscaling_policy: AutoscalingPolicy = AutoscalingPolicy(
            scale_up_window=scale_up_window,
            scale_down_window=scale_down_window,
            scale_to_zero_window=scale_to_zero_window,
        )
        self._autoscaling_policy_sync = SyncAutoscalingPolicy(
            scale_up_window=Duration(seconds=scale_up_window.seconds),
            scale_down_window=Duration(seconds=scale_down_window.seconds),
            scale_to_zero_window=Duration(seconds=scale_to_zero_window.seconds),
        )
        self._region = region
        self._description = description
        self._annotations = annotations
        self._min_replica_count = min_replica_count
        self._max_replica_count = max_replica_count
        self._replica_count = replica_count
        self._accelerator_count = accelerator_count
        self._precision = precision
        self._world_size = world_size
        self._generator_count = generator_count
        self._disaggregated_prefill_count = disaggregated_prefill_count
        self._disaggregated_prefill_world_size = disaggregated_prefill_world_size
        self._max_batch_size = max_batch_size
        self._cluster = cluster
        self._enable_addons = enable_addons
        self._live_merge = live_merge
        self._draft_token_count = draft_token_count
        self._draft_model = draft_model
        self._ngram_speculation_length = ngram_speculation_length
        self._max_peft_batch_size = max_peft_batch_size
        self._kv_cache_memory_pct = kv_cache_memory_pct
        self._enable_session_affinity = enable_session_affinity
        self._direct_route_api_keys = direct_route_api_keys
        self._num_peft_device_cached = num_peft_device_cached
        self._direct_route_type = direct_route_type
        self._direct_route_handle = direct_route_handle
        self._auto_tune = AutoTune()
        self._auto_tune_sync = SyncAutoTune()
        if long_prompt_optimized is not None:
            self._auto_tune.long_prompt_optimized = long_prompt_optimized
            self._auto_tune_sync.long_prompt = long_prompt_optimized

    @property
    def model(self):
        if not self._model.startswith("accounts/fireworks/models/") and "/" not in self._model:
            return f"accounts/fireworks/models/{self._model}"
        return self._model

    @staticmethod
    def _validate_model_for_gpu(model: str, accelerator_type: AcceleratorTypeEnum):
        """
        Models are not always supported on all GPU types. This function checks if the model is supported on a GPU.
        """
        if "qwen3-1p7b" in model:
            supported_accelerators = [
                AcceleratorTypeEnum.NVIDIA_H100_80GB,
                AcceleratorTypeEnum.NVIDIA_H200_141GB,
            ]
            if accelerator_type not in supported_accelerators:
                raise ValueError(
                    f"Qwen3-1p7b is not supported on {accelerator_type}. "
                    f"Please pass one of the following accelerators: {list(map(str, supported_accelerators))} "
                    f"to the LLM constructor using the accelerator_type parameter.\n\n"
                    f"Example:\n"
                    f"    from fireworks import LLM\n"
                    f"    llm = LLM(\n"
                    f'        model="qwen3-1p7b",\n'
                    f'        accelerator_type="NVIDIA_H100_80GB"\n'
                    f"    )"
                )

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._gateway.__aexit__(exc_type, exc_val, exc_tb)

    @property
    def name(self) -> str:
        return self._get_deployment_name()

    @sync_cache
    def _get_deployment_name(self):
        """
        If a name was specified, deployment name will be the specified name.
        Otherwise, the deployment name will be generated from the filename of the caller where this LLM was instantiated.
        """
        if self._name is not None:
            return self._name
        # Get fireworks package path and Python stdlib path
        import fireworks

        package_path = os.path.dirname(fireworks.__file__)
        stdlib_path = sysconfig.get_path("stdlib")

        stack = inspect.stack()
        for frame_info in stack:
            filename = frame_info.filename
            # Skip frames from our package, Python stdlib, other libraries, and internal Python
            if (
                not filename.startswith(package_path)
                and not filename.startswith(stdlib_path)
                and not "/site-packages/" in filename
                and not filename.startswith("<")
            ):
                logger.debug(f"Found caller outside of library code to generate name: {filename}")
                return os.path.basename(filename)
        raise ValueError("No caller found outside of library code")

    def apply(self):
        """
        Like Terraform apply, this will ensure the deployment is ready and return the deployment.
        """
        self._ensure_deployment_ready()
        return self

    def get_time_to_last_token_mean(self) -> Optional[float]:
        """
        Returns the mean time to last token for non-streaming requests. If no metrics are available, returns None.
        """
        if not self.enable_metrics:
            raise ValueError("Metrics are not enabled for this LLM")
        return self._metrics.get_metric_mean("time_to_last_token")

    def is_available_on_serverless(self):
        """
        Checks if the model is available on serverless.
        """
        return self._is_available_on_serverless()

    def deployment_strategy(self) -> Literal["serverless", "on-demand"]:
        return self._deployment_strategy()

    @sync_cache
    def _deployment_strategy(self) -> Literal["serverless", "on-demand"]:
        """
        Determines whether to use serverless or dedicated deployment based on model availability
        and configured deployment type.

        Returns:
            bool: True if a dedicated deployment is needed, False if serverless can be used.
        """
        if self.is_available_on_serverless():
            if self.deployment_type == "serverless":
                logger.debug(f"Model {self.model} is available on serverless, using serverless deployment")
                return "serverless"
            else:
                logger.debug(
                    f"Model {self.model} is available on serverless, but deployment_type is on-demand, continuing to ensure deployment is ready"
                )
                return "on-demand"
        else:
            logger.debug(
                f"Model {self.model} is not available on serverless, continuing to check for existing deployment"
            )
            return "on-demand"

    @sync_cache
    def _is_available_on_serverless(self):
        logger.debug(f"Checking if {self.model} is available on serverless")
        models = self._list_fireworks_models()

        # find model in models
        model = next((m for m in models if m.name == self.model), None)
        if model is None:
            return False
        logger.debug(f"Found model {self.model} on under fireworks account")
        is_serverless = self._is_model_on_serverless_account(model)
        logger.debug(f"Model {self.model} is {'serverless' if is_serverless else 'not serverless'}")
        return is_serverless

    def _list_fireworks_models(self) -> List[Model]:
        """
        Find all models on the fireworks account
        """
        # models = await self._gateway.list_models(parent="accounts/fireworks", include_deployed_model_refs=True)
        return models

    @staticmethod
    def _is_model_on_serverless_account(model: Model) -> bool:
        """
        Check if the model is deployed on a serverless-enabled account.

        Args:
            model: The model object to check

        Returns:
            bool: True if the model is deployed on a supported serverless account, False otherwise
        """
        if model.deployed_model_refs:
            for ref in model.deployed_model_refs:
                if (
                    hasattr(ref, "state")
                    and (ref.state == DeployedModelState.DEPLOYED or ref.state == "DEPLOYED")
                    and hasattr(ref, "deployment")
                    and ref.deployment
                ):
                    # Check if deployment is on a supported account
                    if (
                        ref.deployment.startswith("accounts/fireworks/")
                        or ref.deployment.startswith("accounts/yi-01-ai/")
                        or ref.deployment.startswith("accounts/sentientfoundation/")
                    ):
                        return True
        return False

    def _query_existing_deployment(self) -> Optional[SyncDeployment]:
        """
        Queries all deployments for the same model and accelerator type (if given) and returns the first one.
        """
        deployments = self._gateway.list_deployments_sync(
            filter=f'display_name="{self.name}" AND base_model="{self.model}"'
        )
        deployment = next(iter(deployments), None)
        return deployment

    def _ensure_deployment_ready(self) -> None:
        """
        If a deployment is inferred for this LLM, ensure it's deployed and
        ready. A deployment is required if the model is not available on
        serverless or this LLM has deployment_type="on-demand".

        This method uses a lock to prevent concurrent calls from multiple coroutines.
        """
        deployment_strategy = self.deployment_strategy()
        if deployment_strategy == "serverless":
            return

        deployment = self._query_existing_deployment()

        if deployment is None:
            logger.debug(f"No existing deployment found, creating deployment for {self.model}")
            deployment = SyncDeployment(
                display_name=self.name,
                base_model=self.model,
                autoscaling_policy=self._autoscaling_policy_sync,
            )
            if not isinstance(self._accelerator_type_sync, NotGiven):
                deployment.accelerator_type = self._accelerator_type_sync
            if self._accelerator_count is not None:
                deployment.accelerator_count = self._accelerator_count
            if self._precision is not None:
                deployment.precision = getattr(SyncDeployment.Precision, self._precision)
            if self._world_size is not None:
                deployment.world_size = self._world_size
            if self._generator_count is not None:
                deployment.generator_count = self._generator_count
            if self._region is not None:
                deployment.region = getattr(SyncRegion, self._region)
            if self._description is not None:
                deployment.description = self._description
            if self._annotations is not None:
                deployment.annotations.update(self._annotations)
            if self._min_replica_count is not None:
                deployment.min_replica_count = self._min_replica_count
            if self._max_replica_count is not None:
                deployment.max_replica_count = self._max_replica_count
            if self._replica_count is not None:
                deployment.replica_count = self._replica_count
            if self._disaggregated_prefill_count is not None:
                deployment.disaggregated_prefill_count = self._disaggregated_prefill_count
            if self._disaggregated_prefill_world_size is not None:
                deployment.disaggregated_prefill_world_size = self._disaggregated_prefill_world_size
            if self._max_batch_size is not None:
                deployment.max_batch_size = self._max_batch_size
            if self._cluster is not None:
                deployment.cluster = self._cluster
            if self._enable_addons is not None:
                deployment.enable_addons = self._enable_addons
            if self._live_merge is not None:
                deployment.live_merge = self._live_merge
            if self._draft_token_count is not None:
                deployment.draft_token_count = self._draft_token_count
            if self._draft_model is not None:
                deployment.draft_model = self._draft_model
            if self._ngram_speculation_length is not None:
                deployment.ngram_speculation_length = self._ngram_speculation_length
            if self._max_peft_batch_size is not None:
                deployment.max_peft_batch_size = self._max_peft_batch_size
            if self._kv_cache_memory_pct is not None:
                deployment.kv_cache_memory_pct = self._kv_cache_memory_pct
            if self._enable_session_affinity is not None:
                deployment.enable_session_affinity = self._enable_session_affinity
            if self._direct_route_api_keys is not None:
                deployment.direct_route_api_keys.extend(self._direct_route_api_keys)
            if self._num_peft_device_cached is not None:
                deployment.num_peft_device_cached = self._num_peft_device_cached
            if self._direct_route_type is not None:
                deployment.direct_route_type = getattr(SyncDirectRouteType, self._direct_route_type)
            if self._direct_route_handle is not None:
                deployment.direct_route_handle = self._direct_route_handle
            if self._auto_tune is not None:
                deployment.auto_tune = self._auto_tune_sync
            created_deployment = self._gateway.create_deployment_sync(deployment)
            logger.debug(f"Deployment {created_deployment.name} created, waiting for it to be ready")

            # poll deployment status until it's ready
            start_time = time.time()
            last_log_time = 0
            while created_deployment.state != DeploymentState.READY:
                current_time = time.time()
                # wait for 1 second
                time.sleep(1)
                created_deployment = self._gateway.get_deployment_sync(created_deployment.name)
                if current_time - last_log_time >= 10:
                    elapsed_so_far = current_time - start_time
                    logger.debug(
                        f"Waiting for deployment {created_deployment.name} to be ready, current state: {created_deployment.state}, elapsed time: {elapsed_so_far:.2f}s"
                    )
                    last_log_time = current_time

            total_time = time.time() - start_time
            logger.debug(
                f"Deployment {created_deployment.name} is ready, using deployment (ready in {total_time:.2f} seconds)"
            )
        else:
            logger.debug(f"Deployment {deployment.name} already exists, checking if it needs to be scaled up")

            field_mask = SyncFieldMask()

            # if autoscaling policy is not equal, update it
            if not self._is_autoscaling_policy_equal(deployment):
                logger.debug(
                    f"Updating autoscaling policy for {deployment.name} to "
                    f"{self._autoscaling_policy.scale_up_window.total_seconds()}s up, "
                    f"{self._autoscaling_policy.scale_down_window.total_seconds()}s down, "
                    f"{self._autoscaling_policy.scale_to_zero_window.total_seconds()}s to zero"
                )
                deployment.autoscaling_policy = self._autoscaling_policy_sync
                field_mask.paths.append("autoscaling_policy")

            if deployment.accelerator_type != self._accelerator_type and self._accelerator_type is not NOT_GIVEN:
                raise ValueError(
                    f'Deployment with name "{deployment.name}" has accelerator type {deployment.accelerator_type}, ',
                    f"but the LLM has accelerator type {self._accelerator_type}. You must specify a different name ",
                    f"for this LLM to use a different accelerator type since it is not possible to change the accelerator ",
                    f"type of an existing deployment.",
                )

            if len(field_mask.paths) > 0:
                logger.debug(f"Updating deployment {deployment.name} with {field_mask}")
                start_time = time.time()
                self._gateway.update_deployment_sync(deployment, field_mask)

                # poll until deployment is ready
                while deployment.state != DeploymentState.READY:
                    time.sleep(1)
                    deployment = self._gateway.get_deployment_sync(deployment.name)

                elapsed_time = time.time() - start_time
                logger.debug(f"Deployment update completed in {elapsed_time:.2f} seconds")

            if deployment.replica_count == 0:
                logger.debug(f"Deployment {deployment.name} is not ready, scaling to 1 replica")
                start_time = time.time()
                self._gateway.scale_deployment_sync(deployment.name, 1)

                # Poll until deployment has at least one replica
                last_log_time = 0
                while deployment.replica_count == 0:
                    current_time = time.time()
                    time.sleep(1)
                    deployment = self._gateway.get_deployment_sync(deployment.name)
                    if current_time - last_log_time >= 10:
                        elapsed_so_far = current_time - start_time
                        logger.debug(
                            f"Waiting for deployment {deployment.name} to scale up, current replicas: {deployment.replica_count}, elapsed time: {elapsed_so_far:.2f}s"
                        )
                        last_log_time = current_time

                total_scale_time = time.time() - start_time
                logger.debug(f"Deployment {deployment.name} scaled up in {total_scale_time:.2f} seconds")
            logger.debug(f"Deployment {deployment.name} is ready, using deployment")

    def scale_to_zero(self) -> Optional[Deployment]:
        """
        Sends a request to scale the deployment to 0 replicas but does not wait for it to complete.
        """
        deployment = self.deployment()
        if deployment is None:
            return None
        self._gateway.scale_deployment_sync(deployment.name, 0)
        return deployment

    def scale_to_1_replica(self):
        """
        Scales the deployment to at least 1 replica.
        """
        deployment = self.deployment()
        if deployment is None:
            return
        self._gateway.scale_deployment_sync(deployment.name, 1)

    def _is_autoscaling_policy_equal(self, deployment: Union[Deployment, SyncDeployment]) -> bool:
        if isinstance(deployment, SyncDeployment):
            return (
                deployment.autoscaling_policy.scale_up_window == self._autoscaling_policy_sync.scale_up_window
                and deployment.autoscaling_policy.scale_down_window == self._autoscaling_policy_sync.scale_down_window
                and deployment.autoscaling_policy.scale_to_zero_window
                == self._autoscaling_policy_sync.scale_to_zero_window
            )
        return (
            deployment.autoscaling_policy.scale_up_window == self._autoscaling_policy.scale_up_window
            and deployment.autoscaling_policy.scale_down_window == self._autoscaling_policy.scale_down_window
            and deployment.autoscaling_policy.scale_to_zero_window == self._autoscaling_policy.scale_to_zero_window
        )

    def deployment(self) -> Optional[DeploymentWithIdAndName]:
        return self._deployment()

    @sync_cache
    def _deployment(self) -> Optional[DeploymentWithIdAndName]:
        """
        Queries for any deployment that should exist for this LLM under this
        Fireworks account. If no deployment is found, but should exist, it will
        be created and returned. If no deployment is found and should not exist,
        None is returned.
        """
        self._ensure_deployment_ready()
        deployment = self._query_existing_deployment()
        return DeploymentWithIdAndName(deployment, self._gateway) if deployment else None

    def model_id(self):
        """
        Returns the model ID, which is the model name plus the deployment name
        if it exists. This is used for the "model" arg when calling the model.
        """
        if self.is_available_on_serverless():
            return self.model
        deployment = self.deployment()
        if deployment is None:
            return self.model
        return f"{self.model}#{deployment.id}"

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f"LLM(model={self.model})"

    def fine_tune(
        self,
        name: str,
        dataset: Dataset,
        epochs: Optional[int] = None,
        learning_rate: Optional[float] = None,
        lora_rank: Optional[int] = None,
        jinja_template: Optional[str] = None,
        early_stop: Optional[bool] = None,
        max_context_length: Optional[int] = None,
        base_model_weight_precision: Optional[SupervisedFineTuningJobWeightPrecisionLiteral] = None,
        wandb_config: Optional[WandbConfig] = None,
        evaluation_dataset: Optional[str] = None,
        accelerator_type: Optional[AcceleratorTypeLiteral] = None,
        accelerator_count: Optional[int] = None,
        is_turbo: Optional[bool] = None,
        eval_auto_carveout: Optional[bool] = None,
        region: Optional[RegionLiteral] = None,
        nodes: Optional[int] = None,
        batch_size: Optional[int] = None,
    ):
        """
        Creates a fine-tuning job for this dataset. If the fine-tuning job already exists, it will block until the job is ready.

        Args:
            dataset: The dataset to fine-tune on.
            output_model_name: The name of the output model.
            epochs: The number of epochs to fine-tune for.
            learning_rate: The learning rate to use for fine-tuning.
        """
        if name is None:
            raise ValueError("name is required")
        job = SupervisedFineTuningJob(
            name=name,
            dataset=dataset,
            llm=self,
            epochs=epochs,
            learning_rate=learning_rate,
            lora_rank=lora_rank,
            jinja_template=jinja_template,
            early_stop=early_stop,
            max_context_length=max_context_length,
            base_model_weight_precision=base_model_weight_precision,
            wandb_config=wandb_config,
            evaluation_dataset=evaluation_dataset,
            accelerator_type=accelerator_type,
            accelerator_count=accelerator_count,
            is_turbo=is_turbo,
            eval_auto_carveout=eval_auto_carveout,
            region=region,
            nodes=nodes,
            batch_size=batch_size,
        )
        job = job.sync()
        logger.info(f'Synced fine-tuning job "{name}". See https://fireworks.ai/dashboard/fine-tuning/{job.name}.')
        # poll until job is COMPLETED
        while job.state != JobState.COMPLETED:
            if job.state == JobState.FAILED:
                raise ValueError(f'Fine-tuning job "{name}" failed')
            if job.create_time is not None:
                curr_time = time.time()
                create_time = job.create_time.timestamp()
                delta_seconds = int(curr_time - create_time)
                minutes = delta_seconds // 60
                seconds = delta_seconds % 60
                time_str = f"{seconds}s" if minutes == 0 else f"{minutes}m{seconds}s"
                logger.info(f'Fine-tuning job "{name}" is in state {job.state}. Job was created {time_str} ago.')
            time.sleep(5)
            job = job.get()
            if job is None:
                raise ValueError(f'Fine-tuning job "{name}" not found')
        logger.info(f'Fine-tuning job "{name}" completed')
        if job.output_model is None:
            raise ValueError(f'Fine-tuning job "{name}" did not create an output model')
        return LLM(model=job.output_model)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: gateway/secret.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'gateway/secret.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..google.api import field_behavior_pb2 as google_dot_api_dot_field__behavior__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14gateway/secret.proto\x12\x07gateway\x1a\x1fgoogle/api/field_behavior.proto\"I\n\x06Secret\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xe2\x41\x01\x02\x12\x16\n\x08key_name\x18\x02 \x01(\tB\x04\xe2\x41\x01\x02\x12\x13\n\x05value\x18\x03 \x01(\tB\x04\xe2\x41\x01\x04\"R\n\x13\x43reateSecretRequest\x12\x14\n\x06parent\x18\x01 \x01(\tB\x04\xe2\x41\x01\x02\x12%\n\x06secret\x18\x02 \x01(\x0b\x32\x0f.gateway.SecretB\x04\xe2\x41\x01\x02\"&\n\x10GetSecretRequest\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xe2\x41\x01\x02\"\x8b\x01\n\x12ListSecretsRequest\x12\x14\n\x06parent\x18\x01 \x01(\tB\x04\xe2\x41\x01\x02\x12\x17\n\tpage_size\x18\x02 \x01(\x05\x42\x04\xe2\x41\x01\x01\x12\x18\n\npage_token\x18\x03 \x01(\tB\x04\xe2\x41\x01\x01\x12\x14\n\x06\x66ilter\x18\x04 \x01(\tB\x04\xe2\x41\x01\x01\x12\x16\n\x08order_by\x18\x05 \x01(\tB\x04\xe2\x41\x01\x01\"d\n\x13ListSecretsResponse\x12 \n\x07secrets\x18\x01 \x03(\x0b\x32\x0f.gateway.Secret\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\x12\x12\n\ntotal_size\x18\x03 \x01(\x05\"<\n\x13UpdateSecretRequest\x12%\n\x06secret\x18\x01 \x01(\x0b\x32\x0f.gateway.SecretB\x04\xe2\x41\x01\x02\")\n\x13\x44\x65leteSecretRequest\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xe2\x41\x01\x02\x42\x43ZAgithub.com/fw-ai/fireworks/control_plane/protos/generated/gatewayb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'gateway.secret_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'ZAgithub.com/fw-ai/fireworks/control_plane/protos/generated/gateway'
  _globals['_SECRET'].fields_by_name['name']._loaded_options = None
  _globals['_SECRET'].fields_by_name['name']._serialized_options = b'\342A\001\002'
  _globals['_SECRET'].fields_by_name['key_name']._loaded_options = None
  _globals['_SECRET'].fields_by_name['key_name']._serialized_options = b'\342A\001\002'
  _globals['_SECRET'].fields_by_name['value']._loaded_options = None
  _globals['_SECRET'].fields_by_name['value']._serialized_options = b'\342A\001\004'
  _globals['_CREATESECRETREQUEST'].fields_by_name['parent']._loaded_options = None
  _globals['_CREATESECRETREQUEST'].fields_by_name['parent']._serialized_options = b'\342A\001\002'
  _globals['_CREATESECRETREQUEST'].fields_by_name['secret']._loaded_options = None
  _globals['_CREATESECRETREQUEST'].fields_by_name['secret']._serialized_options = b'\342A\001\002'
  _globals['_GETSECRETREQUEST'].fields_by_name['name']._loaded_options = None
  _globals['_GETSECRETREQUEST'].fields_by_name['name']._serialized_options = b'\342A\001\002'
  _globals['_LISTSECRETSREQUEST'].fields_by_name['parent']._loaded_options = None
  _globals['_LISTSECRETSREQUEST'].fields_by_name['parent']._serialized_options = b'\342A\001\002'
  _globals['_LISTSECRETSREQUEST'].fields_by_name['page_size']._loaded_options = None
  _globals['_LISTSECRETSREQUEST'].fields_by_name['page_size']._serialized_options = b'\342A\001\001'
  _globals['_LISTSECRETSREQUEST'].fields_by_name['page_token']._loaded_options = None
  _globals['_LISTSECRETSREQUEST'].fields_by_name['page_token']._serialized_options = b'\342A\001\001'
  _globals['_LISTSECRETSREQUEST'].fields_by_name['filter']._loaded_options = None
  _globals['_LISTSECRETSREQUEST'].fields_by_name['filter']._serialized_options = b'\342A\001\001'
  _globals['_LISTSECRETSREQUEST'].fields_by_name['order_by']._loaded_options = None
  _globals['_LISTSECRETSREQUEST'].fields_by_name['order_by']._serialized_options = b'\342A\001\001'
  _globals['_UPDATESECRETREQUEST'].fields_by_name['secret']._loaded_options = None
  _globals['_UPDATESECRETREQUEST'].fields_by_name['secret']._serialized_options = b'\342A\001\002'
  _globals['_DELETESECRETREQUEST'].fields_by_name['name']._loaded_options = None
  _globals['_DELETESECRETREQUEST'].fields_by_name['name']._serialized_options = b'\342A\001\002'
  _globals['_SECRET']._serialized_start=66
  _globals['_SECRET']._serialized_end=139
  _globals['_CREATESECRETREQUEST']._serialized_start=141
  _globals['_CREATESECRETREQUEST']._serialized_end=223
  _globals['_GETSECRETREQUEST']._serialized_start=225
  _globals['_GETSECRETREQUEST']._serialized_end=263
  _globals['_LISTSECRETSREQUEST']._serialized_start=266
  _globals['_LISTSECRETSREQUEST']._serialized_end=405
  _globals['_LISTSECRETSRESPONSE']._serialized_start=407
  _globals['_LISTSECRETSRESPONSE']._serialized_end=507
  _globals['_UPDATESECRETREQUEST']._serialized_start=509
  _globals['_UPDATESECRETREQUEST']._serialized_end=569
  _globals['_DELETESECRETREQUEST']._serialized_start=571
  _globals['_DELETESECRETREQUEST']._serialized_end=612
# @@protoc_insertion_point(module_scope)
